#!/usr/bin/env python
# -*- coding:utf-8 -*-

__author__    = 'Mu Yang <emfomy@gmail.com>'
__copyright__ = 'Copyright 2019'

import re
import warnings
import wcwidth

import ply
import ply.lex
import ply.yacc

import ehn.node

################################################################################################################################
# Core
#

EHN_TOKENS_CHAR = {
    'QUOTE':  '"',
    'EQUAL':  '=',
    'COLON':  ':',
    'COMMA':  ',',
    'SLASH':  '/',
    'ULINE':  '_',
    'LPAREN': '(',
    'RPAREN': ')',
    'LBRACE': '{',
    'RBRACE': '}',
    'TILDE':  '~',
}

EHN_TOKENS = [
    'TEXT',
    'NUMBER',
    'COINDEX',
    *EHN_TOKENS_CHAR.keys()
]

class EhnSyntaxError(SyntaxError):

    def __init__(self, *args, pos=None):
        super().__init__(*args)
        self.pos = pos

    def show_pos(self, text):
        return ' '*wcwidth.wcswidth(text[:self.pos]) + '^'

################################################################################################################################
# Lexer
#

class EhnLexer:

    def __init__(self, **kwargs):
        self._lexer = ply.lex.lex(module=self, **kwargs)

    tokens = EHN_TOKENS

    # Define the lexer
    def t_ANY_error(self, t):
        raise EhnSyntaxError('Illegal character ‘{}’ at position {}.'.format(t.value[0], t.lexpos), pos=t.lexpos)
        t.lexer.skip(1)

    # Skip all spaces
    # t_ignore  = ' \t\n\r\f\v'

    # Default state tokens
    t_QUOTE   = r'"'
    t_EQUAL   = r'='
    t_COLON   = r':'
    t_COMMA   = r','
    t_SLASH   = r'/'
    t_ULINE   = r'_'
    t_LPAREN  = r'\('
    t_RPAREN  = r'\)'
    t_LBRACE  = r'{'
    t_RBRACE  = r'}'

    def t_TEXT(self, t):
        r'[A-Za-z0-9\x80-\U0010FFFF|+\-.]+'
        if _isnumber(t.value):
            t.type = 'NUMBER'
        if _is_coindex(t.value):
            t.type = 'COINDEX'
        return t

    def t_TILDE(self, t):
        r'~'
        warnings.warn('‘~’ is deprecated', FutureWarning)
        return t

    # Invoke the lexer
    def __call__(self, data, *args, **kwargs):
        self._lexer.input(data)
        return iter(self._lexer)

################################################################################################################################
# Parser
#

class EhnParser:

    def __init__(self, lexer=None, **kwargs):
        if lexer is not None:
            assert isinstance(lexer, EhnLexer), '{} is not EhnLexer!'.format(lexer)
            self.lexer = lexer
        else:
            self.lexer = EhnLexer()
        self._parser = ply.yacc.yacc(module=self, **kwargs)

    @property
    def _lexer(self):
        return self.lexer._lexer

    tokens = EHN_TOKENS

    # Define the parser
    def p_error(self, t):

        if t is None:
            msg = 'Unexpected ending.'
            pos = None
        else:
            msg = 'Unexpected symbol ‘{}’ at position {}.'.format(t.value, t.lexpos)
            pos = t.lexpos

        syms = []
        for sym in self._parser.action[self._parser.state].keys():
            sym = EHN_TOKENS_CHAR.get(sym, sym)
            if sym == '$end':
                syms.append('‘ENDING’')
            else:
                syms.append('‘{}’'.format(sym))
        if len(syms) > 1:
            syms[-1] = 'or '+syms[-1]

        msg += ' Expecting a {}.'.format(', '.join(syms))
        raise EhnSyntaxError(msg, pos=pos)

    # Object
    def p_expr(self, p):
        '''expr : obj
                | expr COMMA obj'''
        if len(p) == 2:
            p[0] = [p[1]]
        else:
            p[1].append(p[3])
            p[0] = p[1]

    def p_obj(self, p):
        '''obj : entity
               | feature'''
        p[0] = p[1]

    # Entity
    def p_entity_any(self, p):
        '''entityAny : LBRACE RBRACE'''
        p[0] = ehn.node.EhnAnyEntity()

    def p_entity_number(self, p):
        '''entity : LBRACE NUMBER RBRACE'''
        p[0] = ehn.node.EhnNumberEntity(p[2])

    def p_entity_coindex(self, p):
        '''entity : LBRACE COINDEX RBRACE'''
        p[0] = ehn.node.EhnCoindexEntity(p[2])

    def p_entity_name(self, p):
        '''entity : LBRACE QUOTE TEXT QUOTE RBRACE'''
        p[0] = ehn.node.EhnNameEntity(p[3])

    def p_entity_tilde(self, p):
        '''entity : LBRACE TILDE RBRACE'''
        p[0] = ehn.node.EhnTildeEntity()

    def p_entity_normal_open(self, p):
        '''entityOpen : LBRACE TEXT'''
        p[0] = ehn.node.EhnNormalEntity(p[2])

    def p_entity_function_open(self, p):
        '''entityOpen : LBRACE function'''
        p[0] = ehn.node.EhnFunctionEntity(p[2])

    def p_entity_anchor(self, p):
        '''entityAnchor : entityOpen anchor'''
        p[1].anchor = p[2]
        p[0] = p[1]

    def p_entity_feature0(self, p):
        '''entityFeature : entityOpen   COLON feature
                         | entityAnchor COLON feature'''
        p[1].addFeature(p[3])
        p[0] = p[1]

    def p_entity_feature(self, p):
        '''entityFeature : entityFeature COMMA feature'''
        p[1].addFeature(p[3])
        p[0] = p[1]

    def p_entity_close(self, p):
        '''entity : entityOpen    RBRACE
                  | entityAnchor  RBRACE
                  | entityFeature RBRACE'''
        p[0] = p[1]

    # Feature
    def p_feature(self, p):
        '''feature : TEXT EQUAL entity
                   | TEXT EQUAL entityAny
                   | TEXT EQUAL restriction'''
        p[0] = ehn.node.EhnNormalFeature(p[1], p[3])

    def p_function_feature(self, p):
        '''feature : function EQUAL entity
                   | function EQUAL entityAny
                   | function EQUAL restriction'''
        p[0] = ehn.node.EhnFunctionFeature(p[1], p[3])

    # Function
    def p_function_any(self, p):
        '''function : TEXT LPAREN RPAREN'''
        p[0] = ehn.node.EhnFunction(p[1])

    def p_function_restriction(self, p):
        '''function : TEXT LPAREN restriction RPAREN'''
        p[0] = ehn.node.EhnFunction(p[1], p[3])

    def p_function_open(self, p):
        '''functionOpen : TEXT LPAREN entity'''
        p[0] = ehn.node.EhnFunction(p[1], p[3])

    def p_function_argument(self, p):
        '''functionArgument : functionOpen     COMMA entity
                            | functionArgument COMMA entity'''
        p[1].addArgument(p[3])
        p[0] = p[1]

    def p_function_close(self, p):
        '''function : functionOpen     RPAREN
                    | functionArgument RPAREN'''
        p[0] = p[1]

    # Restriction
    def p_restriction(self, p):
        '''restriction : SLASH entity'''
        p[0] = ehn.node.EhnRestriction(p[2])

    def p_restriction_anchor(self, p):
        '''restriction : SLASH entity anchor'''
        p[0] = ehn.node.EhnRestriction(p[2], anchor=p[3])

    # Anchor
    def p_anchor(self, p):
        '''anchor : ULINE COINDEX'''
        p[0] = ehn.node.EhnAnchor(p[2])

    # Invoke the parser
    def __call__(self, data, *args, debug=False, **kwargs):
        if debug:
            print_title(data)
            for tok in self.lexer(data):
                print_info(tok)
        ret = self._parser.parse(data, lexer=self._lexer, *args, debug=debug, **kwargs)
        return ret

################################################################################################################################
# Utility
#

def _isnumber(name):
    try:
        float(name)
        return True
    except:
        return False

def _is_coindex(name):
    return _is_coindex.pattern.match(name)
_is_coindex.pattern = re.compile(r'x[0-9]*')
